@@ -1,5 +1,6 @@
 /*
-  Copyright (C) 2003-2009, 2013-2014 Rocky Bernstein <rocky@gnu.org>
+  Copyright (C) 2003-2009, 2013-2014, 2016-2017 Rocky Bernstein
+  <rocky@gnu.org>
   Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>
 
   This program is free software: you can redistribute it and/or modify
@@ -118,8 +119,6 @@ localtime_r(const time_t *timer, struct tm *result)
 }
 #endif
 
-static const char _rcsid[] = "$Id: iso9660.c,v 1.41 2008/06/25 08:01:54 rocky Exp $";
-
 /* Variables to hold debugger-helping enumerations */
 enum iso_enum1_s     iso_enums1;
 enum iso_flag_enum_s iso_flag_enums;
@@ -206,6 +205,11 @@ iso9660_get_dtime (const iso9660_dtime_t *idr_date, bool b_localtime,
   p_tm->tm_sec    = idr_date->dt_second - idr_date->dt_gmtoff * (15 * 60);
   p_tm->tm_isdst  = -1; /* information not available */
 
+#ifdef HAVE_STRUCT_TM_TM_ZONE
+  /* Initialize everything */
+  p_tm->tm_zone   = 0;
+#endif
+
   /* Recompute tm_wday and tm_yday via mktime. mktime will also renormalize
      date values to account for the timezone offset. */
   {
@@ -273,6 +277,10 @@ iso9660_get_ltime (const iso9660_ltime_t *p_ldate,
 #ifndef HAVE_TM_GMTOFF
   p_tm->tm_sec += p_ldate->lt_gmtoff * (15 * 60);
 #endif
+#ifdef HAVE_STRUCT_TM_TM_ZONE
+  /* Initialize everything */
+  p_tm->tm_zone = 0;
+#endif
 
   /* Recompute tm_wday and tm_yday via mktime. mktime will also renormalize
      date values to account for the timezone offset. */
